// -*- C++ -*-
//
//  This file is part of Siena, a wide-area event notification system.
//  See http://www.cs.colorado.edu/serl/siena/
//
//  Author: Antonio Carzaniga <carzanig@cs.colorado.edu>
//  See the file AUTHORS for full details. 
//
//  Copyright (C) 1998-2002 University of Colorado
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
//  USA, or send email to serl@cs.colorado.edu.
//
//
// $Id: EventServer.cc,v 1.4 2002/11/22 20:57:45 carzanig Exp $
//
#include <sienaserverconf.h>

#include <sys/time.h>
#include <unistd.h>
#include <time.h>

#include <iostream>
#include <sstream>
#include <list>
#include <set>
#include <map>
#include <string>

using namespace std;

#include <siena.h>
#include <siena/Monitor.h>

#include "PThreads.h"
#include "EventServer.h"
#include "CoveringRelations.h"
#include "set_util.h"

#ifdef HAVE_CXX_NAMESPACE
using namespace Siena;
#endif

#ifdef LOG
#undef LOG
#endif
#ifdef LOGENTRY
#undef LOGENTRY
#endif
#define LOG(x)
#define LOGENTRY(x)

template <class T1, class T2>
bool operator != (T1 x, T2 y) { return !(x == y); }

const string EventServer::Version = SIENASERVERVERSION;

static const char * ERR_WritingSenp = "SENP error: to ";

bool EventServer::is_indirect_successor(const Node *a, const Node *b) 
{
    set<Node*>::const_iterator yi;
    for(yi = a->postset.begin(); yi != a->postset.end(); ++yi)
	if (covers((*yi)->filter, b->filter))
	    return true;

    return false;
}

bool EventServer::is_root(const Node *n)
{
    return n->preset.size()==1 && *(n->preset.begin()) == &root;
}

URI EventServer::single_subscriber(const Node *n)
{
    if (n->subscribers.size() != 1) return URI();
    set<URI>::iterator ui;
    ui = neighbors.find(*(n->subscribers.begin()));
    if (ui != neighbors.end()) return *ui;
    return URI();
}

void EventServer::forwards(set<URI> &res, const Node *n)
{
    if (is_root(n)) {
	//
	// if n is a root node, then its forwards is the set of
	// neighbors except for the lonely server subscriber
	//
	res = neighbors;
	if (n->subscribers.size() == 1) {
	    set<URI>::iterator ui;
	    ui = neighbors.find(*(n->subscribers.begin()));
	    if (ui != neighbors.end()) res.erase(*ui);
	}
    } else {
	//
	// if n is not a root node its forwards is the intersection of
	// the single server subscribers of its root predecessors. So
	// if node has a predecessor that is not a root node. its
	// forwards is empty
	//
	res.clear();
	URI res_u;
	set<Node*>::const_iterator xi;
	for(xi = n->preset.begin(); xi != n->preset.end(); ++xi) {
	    if (!is_root(*xi)) return;
	    if (xi == n->preset.begin()) {
		res_u = single_subscriber(*xi);
		if (res_u == URI()) return;
	    } else if (res_u != single_subscriber(*xi)) {
		return;
	    }
	}
	if (res_u != URI()) res.insert(res_u);
    }
}

void EventServer::shutdown()
{
    if (db != NULL) {
	delete(db);
	db = NULL;
    } else {
	set<URI>::iterator ui;
	Request r;
	r[SENP::Method] = SENP::BYE;
	r[SENP::Subscriber] = myuri;
	massage_request(r);
	for(ui = neighbors.begin(); ui != neighbors.end(); ++ui)
	    write_to_siena(*ui, r);
    }
}

void EventServer::db_erase_node(const Node *node)
{
    if (db == NULL) return;
    ostringstream kstr;
    kstr << node->filter;   
    db->erase(Datum(kstr.str()));
}

void EventServer::db_save_node(const Node *node)
{
    if (db == NULL) return;
    ostringstream kstr, vstr;
    set<URI>::const_iterator ui; 
    kstr << node->filter;   
    for(ui = node->subscribers.begin(); ui != node->subscribers.end(); ++ui)
	vstr << (*ui).to_string() << '\n';
    db->set(Datum(kstr.str()), Datum(vstr.str()));
}

static const string Nkey = "neighbors";

void EventServer::db_save_neighbors()
{
    if (db == NULL) return;
    ostringstream vstr;
    set<URI>::const_iterator ui; 
    for(ui = neighbors.begin(); ui != neighbors.end(); ++ui)
	vstr << (*ui).to_string() << endl;
    db->set(Nkey, Datum(vstr.str()));
}

void EventServer::set_db(const char *filename) 
{
    db = new Db(filename);

    neighbors.clear();
    Datum key, nkey, sub;

    key = db->firstkey(); 
    while(!key.is_null()) {
	nkey = db->nextkey(key);
	if (Nkey != key.to_string()) {
	    istringstream fs(string(key.ptr(), key.size()));
	    Filter f;
	    fs >> f;
	    Node * n = new Node(f);

	    sub = db->get(key);
	    istringstream ss(string(sub.ptr(), sub.size()));
	    string u;
	    while(getline(ss, u)) 
		n->subscribers.insert(URI(u));
	    //
	    // now, this can be very expensive... whatever.
	    // ...work in progress...
	    //
	    list<Node *> pred;
	    list<Node *> succ;
	    predecessors(pred, n->filter);
	    successors(succ, n->filter);
	    insert_node_between(n, pred, succ);
	}
	key=nkey;
    }

    key = db->get(Nkey);
    if (!key.is_null()) {
	istringstream ns(string(key.ptr(), key.size()));
	string u;
	while(getline(ns, u)) 
	    neighbors.insert(URI(u));
    }
}

void EventServer::set_monitor(Monitor *m) 
{
    monitor = m;
}

void EventServer::set_log(ostream *lf) 
{
    logf = lf;
}

Request & EventServer::massage_request(Request &r)
{
    //
    // processes a request to be forwarded by this server
    //
    r[SENP::From] = myuri;
    //
    // TTL
    //
    pair<Request::iterator, bool> ri;
    ri = r.insert(Request::value_type(SENP::Ttl, DefaultTTL));
    if (!ri.second) {
	AttributeValue &ttl = (*ri.first).second;
	if (ttl.type() != Siena_integer || ttl.int_value() <= 0) {
	    ttl.int_value() = DefaultTTL;
	} else {
	    --ttl.int_value();
	}
    }
    return r;
}

void EventServer::write_to_siena(const URI &u, Request &r)
{
    //
    // writes a SENP request to Siena. 
    // In the future, this part (i.e., the communication layer) should
    // be handled in a completely different way...
    //
    iostream *s = NULL;
    try {
	r[SENP::To]=u.to_string();
	s = u.open_connection();
	*s << r << endl << flush;
	string reply;
	*s >> reply;
	//
	// should do something with replies ...
	//
	delete(s);
	pending.erase(u);
    } catch (exception &ex) {
	unsigned short a = ++pending[u];
	LOG(ERR_WritingSenp << u.to_string() << " (attempt# " << a << "): " << ex.what() << endl);
	if (s != NULL) delete(s);
	//
	// yeah, whatever!
	//
    }
}

void EventServer::write_to_siena(const URI &u, Request &r, const Event &e)
{
    LOGENTRY("write_to_siena u=" << u.to_string() << " e=" << e << endl);
    //
    // writes a SENP request to Siena. 
    // In the future, this part (i.e., the communication layer) should
    // be handled in a completely different way...
    //
    iostream *s = NULL;
    try {
	if (monitor != NULL) {
	    monitor->notify(myuri, u.to_string());
	}
	r[SENP::To]=u.to_string();
	s = u.open_connection();
	*s << r << e << endl << flush;
	string reply;
	*s >> reply;
	//
	// should do something with replies ...
	//
	delete(s);
	pending.erase(u);
    } catch (exception &ex) {
	unsigned short a = ++pending[u];
	LOG(ERR_WritingSenp << u.to_string() << " (attempt# " << a << "): " << ex.what() << endl);
	if (s != NULL) delete(s);
	//
	// yeah, whatever!
	//
    }
}

void EventServer::write_to_siena(const URI &u, Request &r, const Filter &f)
{
    LOGENTRY("write_to_siena u=" << u.to_string() << " r=" << r << " f=" << f << endl);
    //
    // sends a SENP request to Siena
    // 
    iostream *s = NULL;
    try {
	if (monitor != NULL) {
	    ostringstream ostr; 
	    ostr << f;
	    const string & m =  r[SENP::Method];
	    if (m == "SUB") {
		monitor->subscribe(myuri, u.to_string());
	    } else if (m == "UNS") {
		monitor->unsubscribe(myuri, u.to_string());
	    } else if (m == "PUB") {
		monitor->notify(myuri, u.to_string());
	    }
	}
	s = u.open_connection();
	*s << r << f << endl << flush;
	string reply;
	*s >> reply;
	//
	// should do something with replies ...
	//
	delete(s);
	pending.erase(u);
    } catch (exception &ex) {
	unsigned short a = ++pending[u];
	LOG(ERR_WritingSenp << u.to_string() << " (attempt# " << a << "): " << ex.what() << endl);
	if (s != NULL) delete(s);
	//
	// yeah, whatever!
	//
    }
}

void EventServer::initialize(const string & mu, const string &s)
{
    //
    // initialize the server with a neighbor server
    //
    myuri = mu;
    neighbors.insert(URI(s));
    if (monitor != NULL) {  
	monitor->add_node(myuri, SimpleMonitor::SienaIcon);
	monitor->connect(mu, s);
    }
    db_save_neighbors();
}

void EventServer::initialize(const string & mu)
{
    //
    // initializes the server without neighbor servers
    //
    myuri = mu;
    if (monitor != NULL) {  
	monitor->add_node(myuri, SimpleMonitor::SienaIcon);
    }
}

void EventServer::publish(Request &r, Event &ev) 
{
    LOGENTRY("publish r=" << r << " e=" << ev << endl);
    if (ev.empty()) return;
    URI sender;
    try {
	Request::const_iterator ri = r.find(SENP::From);
	if (ri != r.end() && (*ri).second.type() == Siena_string) 
	    sender = (*ri).second.string_value();
    } catch (BadURI &bu) {
	//
	// I just ignore this exception
	//
    }
    massage_request(r);
    if (r[SENP::Ttl].int_value() == 0) return;
    //
    // tovisit is used as a queue of nodes when walking down the poset
    // this queue implements a breadth-first descent
    // 
    list<Node*>			tovisit;
    set<Node*>			visited;
    list<Node*>::iterator	ni;
    set<Node*>::const_iterator	yi;
    //
    // destinations is the set of subscribers that requested event e
    // (that subscribed for filters that cover e)
    //
    set<URI>				destinations;
    set<URI>::const_iterator		ui;

    for(yi = root.postset.begin(); yi != root.postset.end(); ++yi) 
	if (visited.insert(*yi).second && covers((*yi)->filter, ev))
	    tovisit.push_back(*yi);

    for(ni = tovisit.begin(); ni != tovisit.end(); tovisit.erase(ni++)) {
	destinations += (*ni)->subscribers;
	for(yi = (*ni)->postset.begin(); yi != (*ni)->postset.end(); ++yi) 
	    if (visited.insert(*yi).second && covers((*yi)->filter, ev))
		tovisit.push_back(*yi);
    }
    destinations.erase(sender);
    if (!destinations.empty()) {
	//
	// now forwards the notification
	//
	r[SENP::Method]=SENP::PUB;
	for(ui = destinations.begin(); ui != destinations.end(); ++ui) 
	    write_to_siena(*ui, r, ev);
    }
    clean_up_pending();
}

void EventServer::connect_peer(Request &r) 
{
    LOGENTRY("connect_peer r=" << r << endl);
    //
    // accepts a connection from a peer server. The uri of the peer
    // server is in the "peer" attribute of the SENP request r.
    //
    massage_request(r);
    if (r[SENP::Ttl].int_value() == 0) return;

    set<URI>::iterator new_uri 
	= neighbors.insert(URI(r[SENP::Peer].string_value())).first;
    //
    // after inserting the peer server in the set of neighbors,
    // forward the set of ``root'' filters to the new peer.
    //
    r[SENP::Method]=SENP::SUB;
    r[SENP::Subscriber]=myuri;
    set<Node*>::iterator yi;
    for(yi = root.postset.begin(); yi != root.postset.end(); ++yi)
	write_to_siena((*new_uri), r, (*yi)->filter);
    db_save_neighbors();
}
 
bool EventServer::exists_in_predecessors(list<Node*> & pred, 
					 const Filter &f, 
					 const URI &U)
{
    //
    // this function starts from root and walks down the poset
    // looking for the first subscription that covers f that contains
    // U (the new subscriber). If such a subscription is found, the
    // result will be TRUE and pred will contain the (one) pointer to
    // that node. Otherwise, the result will be false and pred will
    // contain the list of the immediate predecessors of f. 
    //
    set<Node*>		visited;
    list<Node*>::iterator	ni;
    set<Node*>::iterator	yi;
    bool			found_lower;
    
    pred.clear();
    pred.push_back(&root);

    ni = pred.begin();
    while(ni != pred.end()) {
	found_lower = false;
	for(yi = (*ni)->postset.begin(); yi != (*ni)->postset.end(); ++yi) {
	    if (covers((*yi)->filter, f)) {
		found_lower = true;
		if (visited.insert(*yi).second) {
		    if ((*yi)->subscribers.find(U) != (*yi)->subscribers.end()) {
			pred.clear(); 
			pred.push_back(*yi);
			return true;
		    }
		    pred.push_back(*yi);
		}
	    }
	}
	if (found_lower) {
	    pred.erase(ni++);
	} else {
	    ++ni;
	}
    }
    return false;
}

void EventServer::successors(list<Node*> & succ, const Filter & f)
{
    //
    // computes (in succ) the set of immediate successors of a given
    // filter f
    //
    set<Node*>				visited;
    list<Node*>::iterator		ni;
    set<Node*>::const_iterator		yi;

    succ.clear();
    succ.push_back(&root);

    for(ni = succ.begin(); ni != succ.end(); succ.erase(ni++))
	for(yi = (*ni)->postset.begin(); yi != (*ni)->postset.end(); ++yi) 
	    if (visited.insert(*yi).second)
		if (covers(f, (*yi)->filter)) {
		    succ.push_front(*yi);
		} else {
		    succ.push_back(*yi);
		}
}

void EventServer::predecessors(list<Node*> & pred, const Filter & f)
{
    //
    // computes (in pred) the set of immediate predecessors of a given
    // filter f
    //
    set<Node*>				visited;
    list<Node*>::iterator		ni;
    set<Node*>::const_iterator		yi;
    bool				found_lower;

    pred.clear();
    pred.push_back(&root);
    
    ni = pred.begin();
    while(ni != pred.end()) {
	found_lower = false;
	for(yi = (*ni)->postset.begin(); yi != (*ni)->postset.end(); ++yi)
	    if (covers((*yi)->filter, f)) {
		found_lower = true;
		if (visited.insert(*yi).second)
		    pred.push_back(*yi);
	    }
	if (found_lower) {
	    pred.erase(ni++);
	} else {
	    ++ni;
	}
    }
}

void EventServer::disconnect_nodes(Node * x, Node * y) 
{
    if (x->postset.erase(y) > 0) 
	y->preset.erase(x);
}

void EventServer::connect_nodes(Node * x, Node * y) 
{
    if (x->postset.insert(y).second) 
	y->preset.insert(x);
}

void EventServer::erase_node(Node * node)
{
    //
    // this basically undoes insert_node_between
    //
    set<Node*>::iterator xi;
    set<Node*>::iterator yi;
    //
    // 1. disconnect every successor from node
    //
    for(yi = node->postset.begin(); yi != node->postset.end(); ++yi)
	(*yi)->preset.erase(node);
    //
    // 2. disconnect predecessors from node thereby reconnecting
    // predecessors to successors. A predecessor X is re-connected to
    // a successor Y only if X doesn't have an immediate successor X'
    // that covers Y.
    //
    for(xi = node->preset.begin(); xi != node->preset.end(); ++xi) {
	(*xi)->postset.erase(node);
	for(yi = node->postset.begin(); yi != node->postset.end(); ++yi) 
	    if (!is_indirect_successor(*xi, *yi))
		connect_nodes(*xi, *yi);
    }
    db_erase_node(node);
    delete(node);
}

void EventServer::insert_node_between(Node * node,
				      list<Node*> & pred,
				      list<Node*> & succ)
{
    //
    // add links with predecessors and successors, and at the same
    // time remove old links that become indirect because of the new
    // node (i.e., those between the successors and the predecessors
    // of the new node)
    //
    list<Node*>::iterator xi;
    list<Node*>::iterator yi;
    for(xi = pred.begin(); xi != pred.end(); ++xi) {
	connect_nodes(*xi, node);
	for(yi = succ.begin(); yi != succ.end(); ++yi) 
	    disconnect_nodes(*xi, *yi);
    }
    for(yi = succ.begin(); yi != succ.end(); ++yi) 
	connect_nodes(node, *yi);
}

void EventServer::clear_subtree(const Node *node, const URI & U)
{
    //
    // removes U from subscribers for all the nodes of the sub-poset
    // rooted in node, not including node itself.
    //
    set<Node*>			visited;
    list<Node*>			tovisit;
    list<Node*>::iterator	ni;
    set<Node*>::iterator	yi;
    //
    // initializes tovisit with node's successors
    // 
    for(yi = node->postset.begin(); yi != node->postset.end(); ++yi)
	if(visited.insert(*yi).second) 
	    tovisit.push_back(*yi);

    for(ni = tovisit.begin(); ni != tovisit.end(); tovisit.erase(ni++)) {
	for(yi = (*ni)->postset.begin(); yi != (*ni)->postset.end(); ++yi)
	    if(visited.insert(*yi).second) 
		tovisit.push_back(*yi);
	if ((*ni)->subscribers.erase(U) > 0) {
	    if ((*ni)->subscribers.empty()) {
		//
		// erasing *nodei from the poset is safe here since
		// nodei iterates on a different list: nqueue
		//
		erase_node(*ni);
	    } else {
		db_save_node(*ni);
	    }
	}
    }
}

void EventServer::clean_up_pending()
{
    if(cleaning_pending) return;
    cleaning_pending = true;
    map<URI, unsigned short>::iterator pi;
    Request r; 
    massage_request(r);
    pi = pending.begin(); 
    while(pi != pending.end())
	if ((*pi).second >= max_attempts) {
	    list<Node*>			succ;
	    set<Node*>::iterator	yi;
	    
	    for(yi = root.postset.begin(); yi != root.postset.end(); ++yi)
		succ.push_back(*yi);
	    
	    unsubscribe_from_succ(r, succ, (*pi).first);
	    pending.erase(pi++);
	} else {
	    ++pi;
	}
    cleaning_pending = false;
}

void EventServer::subscribe(Request &r, Filter &f) 
{
    LOGENTRY("subscribe r=" << r << " f=" << f << endl);
    if (f.empty()) return;
    URI U(r[SENP::Subscriber].string_value());
 
    list<Node*> pred;
    set<URI> Z;

    if (exists_in_predecessors(pred,f,U)) {
	//
	// U is already subscribed for a subscription that covers f
	//
	return;
    }
    Node *sub_node;
    if (pred.size() == 1 && covers(f,(*pred.begin())->filter)) {
	//
	// a subscription f is already present, I just add U. If
	// *sub_node is a root node that has a single_subscriber
	// (before inserting U) the subscription is forwarded to the
	// single_subscriber
	// 
	sub_node = *pred.begin();
	if (is_root(sub_node)) {
	    URI u = single_subscriber(sub_node);
	    if (u != URI()) Z.insert(u);
	}
	sub_node->subscribers.insert(U);
    } else {
	//
	// the subscription is NEW! --> create and connect the new node
	//
	sub_node = new Node(f);
	sub_node->subscribers.insert(U);
	list<Node*> succ;
	successors(succ, f);
	insert_node_between(sub_node, pred, succ);
	//
	// forwards to neighbors appropriately (uses forwards())
	//
	forwards(Z, sub_node);
    }
    if (!Z.empty()) {
	massage_request(r);
	if (r[SENP::Ttl].int_value() > 0) {
	    r[SENP::Subscriber] = myuri;
	    for(set<URI>::iterator ui = Z.begin(); ui != Z.end(); ++ui)
		write_to_siena(*ui, r, f);
	}
    }
    clear_subtree(sub_node, U);
    db_save_node(sub_node);
    clean_up_pending();
    return;
}

void EventServer::unsubscribe(Request &r, Filter &f) 
{
    LOGENTRY("unsubscribe r=" << r << " f=" << f << endl);
    URI U(r[SENP::Subscriber].string_value());
    list<Node*> pred;
    list<Node*> succ;

    if (exists_in_predecessors(pred, f, U)) {
	if (!covers(f, (*pred.begin())->filter)) {
	    //
	    // U has subscribed for a subscription that covers
	    // (strictly) f, so this unsubscription has no effect.
	    //
	    return;
	} else {
	    //
	    // U has subscribed for f (exactly)
	    // 
	    succ.push_front(*pred.begin());
	}
    } else {
	//
	// U has not subscribed for anything ``above'' f It might have
	// subscribed for some f' below f, so we need to look for those f'
	//
	successors(succ, f);
    }
    //
    // here, U has not been found in any subscription covering f, so I
    // can proceed to see if it exists in any one of the successors of
    // f. In that case, U will be removed from all of them. 
    //
    unsubscribe_from_succ(r, succ, U);
    clean_up_pending();
}

void EventServer::unsubscribe_from_succ(Request & r, 
					list<Node*> & succ, 
					const URI & U)
{
    //
    // this is the actual unsubscription procedure, it eliminates all
    // the subscriptions of U starting from the set of nodes succ.
    //
    set<Node*>			visited;
    set<Node*>			to_resub;
    list<Node*>::iterator 	qi;
    set<URI> 			Z;
    set<URI>::iterator		ui;
    set<Node*>::iterator	yi;
    set<Node*>::iterator	ni;

    massage_request(r);
    r[SENP::Subscriber] = myuri;

    Request resub = r;
    resub[SENP::Method] = SENP::SUB;
    //
    // reset method to UNS, in case this is a follow up of a BYE request
    //
    r[SENP::Method] = SENP::UNS;

    for(qi = succ.begin(); qi != succ.end(); succ.erase(qi++)) {
	ui = (*qi)->subscribers.find(U);
	if (ui != (*qi)->subscribers.end()) {
	    Z.clear();
	    //
	    // this might be done more efficiently ... maybe...
	    // ...work in progress...
	    //
	    if ((*qi)->subscribers.size() == 1) {
		forwards(Z, *qi);
	    }
	    (*qi)->subscribers.erase(ui);
	    if (is_root(*qi)) {
		URI u = single_subscriber(*qi);
		if (u != URI()) Z.insert(u);
	    }
	    if (!Z.empty() && r[SENP::Ttl].int_value() > 0) {
		for(ui = Z.begin(); ui != Z.end(); ++ui)
		    write_to_siena(*ui, r, (*qi)->filter);
		//
		// now figures out which of the successors of this
		// node must be re-subscribed: every succ goes to the
		// intersection of its natural forwards() (after
		// removing *qi) and Z.
		//
		to_resub = (*qi)->postset;
		    
		if ((*qi)->subscribers.empty()) erase_node(*qi);
		    
		for(ni = to_resub.begin(); ni != to_resub.end(); ++ni) {
		    set<URI> to_rs;
		    forwards(to_rs, *ni);
		    to_rs &= Z;
		    for(ui = to_rs.begin(); ui != to_rs.end(); ++ui) 
			write_to_siena(*ui, resub, (*ni)->filter);
		}
	    } else {
		if ((*qi)->subscribers.empty()) erase_node(*qi);
	    }
	} else {
	    //
	    // U is not in *qi, therefore visit its successors
	    //
	    for(yi = (*qi)->postset.begin(); yi != (*qi)->postset.end(); ++yi)
		if (visited.insert(*yi).second)
		    succ.push_back(*yi);
	}
    }
}

void EventServer::unsubscribe_all(Request &r)
{
    //
    // eliminates the subscriber from all the subscriptions.
    // Basically, the idea is to this scans the poset. The search is
    // pruned whenever U is found among the subscribers (of the
    // current node). Notice the case in which U is a peer server.
    //
    URI U(r[SENP::Subscriber].string_value());

    neighbors.erase(U);

    list<Node*>			succ;
    set<Node*>::iterator	yi;

    for(yi = root.postset.begin(); yi != root.postset.end(); ++yi)
	succ.push_back(*yi);

    unsubscribe_from_succ(r, succ, U);
    clean_up_pending();
}

void EventServer::advertise(Request &r, Filter &f) 
{
    LOGENTRY("advertise r=" << r << " f=" << f << endl);
    return;
}

void EventServer::unadvertise(Request &r, Filter &f) 
{
    LOGENTRY("unadvertise r=" << r << " f=" << f << endl);
    return;
}

void EventServer::print_poset(ostream &os) 
{
    //
    // tovisit is used as a queue of nodes when walking down the poset
    // this queue implements a breadth-first descent
    // 
    list<Node*>			tovisit;
    set<Node*>			visited;
    list<Node*>::iterator	ni;
    set<URI>::const_iterator	ui;
    set<Node*>::const_iterator	zi;

    os << "Siena Server v." << Version << endl;
    if (!neighbors.empty()) {
	os << " <->";
	for(ui = neighbors.begin(); ui != neighbors.end(); ++ui)
	    os << ' ' << (*ui).to_string();
	os << endl;
    }

    for(zi = root.postset.begin(); zi != root.postset.end(); ++zi)
	if (visited.insert(*zi).second)
	    tovisit.push_back(*zi);

    for(ni = tovisit.begin(); ni != tovisit.end(); tovisit.erase(ni++)) {
	os << "------- " << *ni << endl;
	os << (*ni)->filter << endl;
	os << "  >>";
	for(ui = (*ni)->subscribers.begin(); 
	    ui != (*ni)->subscribers.end(); ++ui)
	    os << ' ' << (*ui).to_string();
	os << endl << "  ^^";
	for(zi = (*ni)->preset.begin(); zi != (*ni)->preset.end(); ++zi)
	    if (*zi == &root) {
		os << " root";
	    } else {
		os << ' ' << *zi;
	    }
	os << endl;
	os << "  vv";
	for(zi = (*ni)->postset.begin(); zi != (*ni)->postset.end(); ++zi) {
	    if (visited.insert(*zi).second)
		tovisit.push_back(*zi);
	    os << ' ' << *zi;
	}
	os << endl;
    }
}

unsigned short EventServer::attempts() { 
    return max_attempts; 
}

unsigned short EventServer::attempts(unsigned short ma) { 
    unsigned short res = max_attempts;
    max_attempts = ma;
    return res;
}
